#!/usr/bin/env python3
from samshee.samplesheetv2 import SampleSheetV2, read_samplesheetv2
from samshee.sectionedsheet import read_sectionedsheet, parse_sectionedsheet_from_json
from samshee.validation import validate
import sys
import argparse
import json
from pathlib import Path
from io import TextIOBase

parser = argparse.ArgumentParser()

parser.add_argument(
    "infile",
    help='input sample sheet. Use "-" for stdin.',
)
parser.add_argument(
    "--input-format",
    help="expect the input file to be in this format (possible values: sectioned, json) (v1 sheets can be read as sectioned sheets but are not validated by default!)",
    default="sectioned",
)
parser.add_argument(
    "--output-format",
    help="print the read sheet in this format (possible values: sectioned, v2, json).",
    default="v2",
)
parser.add_argument(
    "--schema",
    help='specifies an additional schema (other than those required by the input or output formats) in json format, e.g. \'{"required": ["Header", "Reads"]}\'. Use "$ref" for linking to predefined external or samshee-internal schemas, e.g., \'{"$ref": "urn:samshee:illuminav2/v1"}\'. This argument can be given more than once.',
    default=[],
    action="append",
    type=str,
)

args = parser.parse_args()

if args.infile == "-":
    infile = sys.stdin  # is a open TextIOWrapper
else:
    infile = Path(args.infile)  # needs to be opened

try:
    if args.input_format == "json":
        if isinstance(infile, TextIOBase):
            jstr = infile.read()
        else:
            with open(infile, "r") as fh:
                jstr = fh.read()
        sheet = parse_sectionedsheet_from_json(jstr)
    elif args.input_format == "sectioned":
        sheet = read_sectionedsheet(infile)
    else:
        print(
            f'Unknown input format: {args.input_format}. Must be json or sectioned. Use "sectioned" for any v1 or v2 sample sheets.',
            file=sys.stderr,
        )
        sys.exit(3)
except Exception as exc:
    print(f"Error reading sheet: {exc}", file=sys.stderr)
    sys.exit(1)

try:
    validate(sheet, validation=[json.loads(schema) for schema in args.schema])
except Exception as exc:
    print(f"Validation failed: {exc}", file=sys.stderr)
    sys.exit(2)


try:
    if args.output_format == "json":
        print(sheet.to_json(pretty=True))
    elif args.output_format == "v2":
        print(SampleSheetV2(sheet))
    elif args.output_format == "sectioned":
        print(sheet)
    else:
        print(
            f"Unknown output format: {args.output_format}. Must be json, v2, or sectioned.",
            file=sys.stderr,
        )
        sys.exit(3)
except Exception as exc:
    print(f"failed writing output: {exc}", file=sys.stderr)
    sys.exit(4)



#!/usr/bin/env python3
import pytest
from samshee.sectionedsheet import SectionedSheet, read_sectionedsheet, parse_array
import tempfile
from io import StringIO
from pathlib import Path


def test_can_read_from_open_file():
    fh = StringIO(
        """
[Header],
FileFormatVersion,2
RunName,p123_A_scGEX_scCSP_Novaseq
InstrumentPlatform,NovaSeq
IndexOrientation,Forward

[Reads]
Read1Cycles,28
Read2Cycles,90
Index1Cycles,10
Index2Cycles,10
"""
    )
    read_sectionedsheet(fh)


def test_can_read_from_path():
    tempdir = tempfile.TemporaryDirectory()
    fname = Path(tempdir.name) / "sheet.csv"
    with fname.open("w") as fh:
        fh.write(
            """
[Header],
FileFormatVersion,2
RunName,p123_A_scGEX_scCSP_Novaseq
InstrumentPlatform,NovaSeq
IndexOrientation,Forward

[Reads]
Read1Cycles,28
Read2Cycles,90
Index1Cycles,10
Index2Cycles,10
"""
        )

    read_sectionedsheet(fname)


def test_can_read_binary_from_path():
    tempdir = tempfile.TemporaryDirectory()
    fname = Path(tempdir.name) / "sheet.csv"
    with fname.open("w") as fh:
        fh.write(
            """
[Header],
FileFormatVersion,2
RunName,p123_A_scGEX_scCSP_Novaseq
InstrumentPlatform,NovaSeq
IndexOrientation,Forward

[Reads]
Read1Cycles,28
Read2Cycles,90
Index1Cycles,10
Index2Cycles,10
"""
        )

    with fname.open("rb") as fh:
        read_sectionedsheet(fh)

def test_can_read_array():
    s = """
[Reads]
28,
90,
10,,
10
"""
    a = parse_array(s)
    assert len(a) == 4
    assert a[0] == 28
    assert a[1] == 90
    assert a[2] == 10
    assert a[3] == 10



import re
import json
from collections import OrderedDict
from typing import Optional, Callable, cast

from samshee.sectionedsheet import (
    Settings,
    SectionedSheet,
    Section,
    Data,
    read_sectionedsheet,
    parse_sectionedsheet_from_json,
)
from samshee.validation import (
    validate,
    illuminasamplesheetv2logic,
)
from samshee.validation import registry as samsheeschemaregistry


class SampleSheetV2:
    """A class that represents an illumina Sample Sheet v2.
    This is always constructed from a SectionedSheet that typically has been validated against a set of rules.
    """

    def __init__(
        self,
        secsheet: SectionedSheet = SectionedSheet(),
        validation: list[Callable | dict] = [
            {"$ref": "urn:samshee:illuminav2/v1"},
            illuminasamplesheetv2logic,
        ],
        registry=samsheeschemaregistry,
    ) -> None:
        """Parsing from"""
        validate(cast(SectionedSheet, secsheet), validation, registry=registry)
        self.validation = validation
        self.registry = registry

        def secname(k: str) -> str:
            secsel = re.compile("^(.*)_(Settings|Data)$")
            m = re.match(secsel, k)
            if m is not None:
                return m.group(1)
            return k

        self.applications: OrderedDict[str, dict[str, Section]] = OrderedDict()
        for key in secsheet.keys():
            sectionname = secname(key)
            if key.endswith("_Settings"):
                if sectionname not in self.applications:
                    self.applications[sectionname] = dict[str, Settings | Data]()
                self.applications[sectionname]["settings"] = cast(
                    Settings, secsheet[key]
                )
            elif key.endswith("_Data"):
                if sectionname not in self.applications:
                    self.applications[sectionname] = dict()
                self.applications[sectionname]["data"] = secsheet[key]
            elif key == "Header":
                self.header = Settings(secsheet["Header"])
            elif key == "Reads":
                self.reads = secsheet["Reads"]

    def to_sectionedsheet(self, validate_schema=True) -> SectionedSheet:
        """Constructs a SectionedSheet, unless validate_schema is False, the sheet is revalidated"""
        res = SectionedSheet(OrderedDict())
        if "header" in self.__dict__.keys():
            res["Header"] = self.header
        if "reads" in self.__dict__.keys():
            res["Reads"] = self.reads
        for appname, app in self.applications.items():
            if "settings" in app:
                res[appname + "_Settings"] = app["settings"]
            if "data" in app:
                res[appname + "_Data"] = app["data"]
        if validate_schema:
            validate(res, self.validation, self.registry)
        return res

    def __str__(self) -> str:
        """returns a string representation of the sample sheet (adheres to the usual sample sheet format)"""
        return str(self.to_sectionedsheet())

    def write(self, filehandle) -> None:
        """writes a string representation of the sample sheet (adheres to the usual sample sheet format) to a file"""
        return self.to_sectionedsheet().write(filehandle)

    def to_json(self, pretty=False) -> str:
        return self.to_sectionedsheet().to_json(pretty)


def read_samplesheetv2(
    fromfile,
    validation=[{"$ref": "urn:samshee:illuminav2/v1"}, illuminasamplesheetv2logic],
    registry=samsheeschemaregistry,
) -> SampleSheetV2:
    """reads a SampleSheetv2 from a file by first parsing it as a SectionedSheet and then validating it against the standard schemata"""
    return SampleSheetV2(
        read_sectionedsheet(fromfile), validation=validation, registry=registry
    )


def parse_samplesheetv2_from_json(
    jsonstr: str,
    validation=[{"$ref": "urn:samshee:illuminav2/v1"}, illuminasamplesheetv2logic],
    registry=samsheeschemaregistry,
) -> SampleSheetV2:
    """parses a SampleSheetv2 from a json string by first parsing it as a SectionedSheet and then validating it against the standard schemata"""
    return SampleSheetV2(parse_sectionedsheet_from_json(jsonstr), validation)


def parse_samplesheetv2_from_object(
    obj,
    validation=[{"$ref": "urn:samshee:illuminav2/v1"}, illuminasamplesheetv2logic],
    registry=samsheeschemaregistry,
) -> SampleSheetV2:
    """constructs a SampleSheetv2 from a object (dict) by first constructing a SectionedSheet from it and then validating it against the standard schemata"""
    return parse_samplesheetv2_from_json(json.dumps(obj))
